#include "planner.hpp"
#include "action.hpp"
#include "global_world_state.hpp"
#include "godot_cpp/templates/pair.hpp"
#include "utils/godot_macros.h"
#include <godot_cpp/templates/hash_set.hpp>
#include <godot_cpp/templates/hashfuncs.hpp>
#include <godot_cpp/variant/vector3.hpp>

namespace godot::goap {
void Planner::_bind_methods() {
#define CLASSNAME Planner
    GDPROPERTY_HINTED(actions, Variant::OBJECT, PROPERTY_HINT_ARRAY_TYPE, "Action");
}

void Planner::_ready() {
    this->global_world_state = GlobalWorldState::get_singleton();
}

void Planner::_process(double delta_time) {
    this->cached_world_state.clear();
}

Vector<State> Planner::make_plan(Goal *goal) {
    HashMap<PlannerNode, int, PlannerNodeHasher> open{};
    open.insert(PlannerNode{{}, goal->goal_state, nullptr}, 0);
    HashMap<PlannerNode, Ref<Action>, PlannerNodeHasher> from{};
    HashMap<PlannerNode, float, PlannerNodeHasher> score_to_start{};
    HashMap<PlannerNode, float, PlannerNodeHasher> heuristic_score{};

    PlannerNode current;
    while(!open.is_empty()) {
        int current_weight = 0;
        for(KeyValue<PlannerNode, int> &kvp : open) {
            if(kvp.value > current_weight) {
                current = kvp.key;
                current_weight = kvp.value;
            }
        }
    }
}

Variant Planner::get_world_property(StringName prop_key) {
    if(world_state_override.has(prop_key))
        return world_state_override.get(prop_key);
    if(prop_key.begins_with("g_"))
        return this->global_world_state->get_world_property(prop_key);
    if(this->cached_world_state.has(prop_key))
        return this->cached_world_state[prop_key];
    if(this->has_method("get_" + prop_key)) {
        Variant val = this->call(prop_key);
        this->cached_world_state[prop_key] = val;
        return val;
    }
    return nullptr;
}

bool Planner::can_do(Ref<Action> action) {
    for(WorldProperty &prop : action->context_prerequisites) {
        if(this->get_world_property(prop.key) != prop.value)
            return false;
    }
    return true;
}

Vector<Ref<Action>> Planner::find_actions_satisfying(WorldState requirements) {
    Vector<Ref<Action>> found_actions{};
    for(Ref<Action> &act : this->actions) {
        for(WorldProperty &prop : requirements) {
            if(act->effects.has(prop.key)
            && act->effects.get(prop.key) == prop.value
            && this->can_do(act))
                found_actions.push_back(act);
        }
    }
    return found_actions;
}

void Planner::set_actions(Array value) {
    this->actions.clear();
    for(size_t i{0}; i < value.size(); ++i) {
        Ref<Action> act = value[i];
        if(act.is_valid())
            this->actions.push_back(value[i]);
    }
}

Array Planner::get_actions() const {
    Array array{};
    for(Ref<Action> const &act : this->actions) {
        array.push_back(act);
    }
    return array;
}
}
