#ifndef GOAP_PLANNER_HPP
#define GOAP_PLANNER_HPP

#include "action.hpp"
#include "godot_cpp/classes/object.hpp"
#include "godot_cpp/variant/variant.hpp"
#include <godot_cpp/classes/node.hpp>
#include <godot_cpp/classes/resource.hpp>
#include <godot_cpp/templates/vector.hpp>
#include <godot_cpp/variant/string_name.hpp>

namespace godot {
class CharacterActor;

namespace goap {
class GlobalWorldState;

struct State {
    static State new_move_to(Vector3 location);
    static State new_animate(StringName animation);
    static State new_activate(Node *node);
    enum Type {
        STATE_MOVE_TO,
        STATE_ANIMATE,
        STATE_ACTIVATE
    };
    State::Type type;
    union {
        Vector3 move_to;
        StringName animate;
        Node *activate;
    };
};

class Goal : public Resource {
    GDCLASS(Goal, Resource);
    static void _bind_methods();
public:
    WorldState goal_state{};
};

class Planner : public Node {
    GDCLASS(Planner, Node);
    static void _bind_methods();
public:
    virtual void _ready() override;
    virtual void _process(double delta_time) override;

    Vector<State> make_plan(Goal *goal);

    Variant get_world_property(StringName prop_key);

    bool can_do(Ref<Action> action);
    Vector<Ref<Action>> find_actions_satisfying(WorldState requirements);

    void set_actions(Array actions);
    Array get_actions() const;
private:
    CharacterActor *actor{nullptr};
    WorldState world_state_override{};
    WorldState cached_world_state{};
    GlobalWorldState *global_world_state{nullptr};

    Vector<Ref<Action>> actions{};
};

struct PlannerNode {
    WorldState state;
    WorldState open_requirements;
    Ref<Action> last_edge;

    int heuristic_score(Ref<Action> action) {
        int score{0};
        return score;
    }
};

struct PlannerNodeHasher {
    static _FORCE_INLINE_ uint32_t hash(godot::goap::PlannerNode const &node) {
        VariantHasher variant_hasher{};
        Variant a{1};
        variant_hasher.hash(a);
    }
};

static _FORCE_INLINE_ bool operator==(PlannerNode const& lhs, PlannerNode const& rhs) {
    return true;
}
}
}


#endif // !GOAP_PLANNER_HPP
